#ifndef DEVICE_SHADOW_H
#define DEVICE_SHADOW_H

#include <cvector.h>
#include <dbg.h>
#include <emqc_base.h>
#include <json-c/json.h>
#include <stdbool.h>

// TODO
typedef enum { 
    INT, 
    STR, 
    ARR 
} TAG;
// typedef enum {INT, STR, ARR} TYPE;
typedef enum { 
    REPORTED = 0,
    DESIRED = 1 
} emqc_ds_state;

typedef int (*callback)(void *);

struct emqc_ds {
    char *   sub_topic;
    char *   pub_topic;
    emqc *   handle;
    callback cb[3];
};

typedef struct emqc_ds   emqc_ds;
typedef struct attr      attr;
typedef struct serv_resp serv_resp;
typedef struct serv_req  serv_req;

typedef struct attr_in {
    TAG tag;
    union {
        int   val_int;
        char *val_str;
    };
} attr_in;

struct attr {
    TAG   tag;
    char *key;
    union {
        int      val_int;
        char *   val_str;
        attr_in *val_arr;
    };
    long timestamp;
};

/**
 *  @ brief Create a new emqc device shadow client instance.
 *  @ ret : emqc client instance.
 */
emqc_ds *emqc_ds_new();

/**
 *  @ brief Initialize emqc device shadow client instance.
 *  @ param handle: emqc handle.
 *  @ param host : hostname or ip address of the broker to connect to.
 *  @ param port : network port to connect to.  Usually 1883.
 *  @ ret : error code.
 */
emqc_ds *emqc_ds_init(emqc_ds *handle, char *host, int port);

/**
 *  @ brief Use to free memory associated with a emqc client instance.
 *  @ param handle : emqc client instance.
 *  @ ret :
 */
void emqc_ds_destroy(emqc_ds *handle);

/**
 *  @ brief Connect to an mqtt broker.
 *  @ param handle : an emqc device shadow client instance.
 *  @ param keep_alive_interval : keep alive interval.
 *  @ param clean_session : clean session.
 *  @ ret : error code.
 */
EMQC_ERR_CODE emqc_ds_connect(emqc_ds *handle, int keep_alive_interval,
                              int clean_session);

/**
 *  @ brief Set host.
 *  @ param handle : an emqc device shadow client instance.
 *  @ param host : host.
 */
void emqc_ds_set_host(emqc_ds *handle, char *host);

/**
 *  @ brief Set port.
 *  @ param handle : an emqc device shadow client instance.
 *  @ param port : port.
 */
void emqc_ds_set_port(emqc_ds *handle, int port);

/**
 *  @ brief Disconnect from a broker.
 *  @ param handle : an emqc device shadow client instance.
 *  @ ret : error code.
 */
EMQC_ERR_CODE emqc_ds_disconnect(emqc_ds *handle);

/**
 *  @ brief Set get callback.
 *  @ param handle : an emqc device shadow client instance.
 *  @ param func : a callback function for receive data.
 */
void emqc_ds_set_get_callback(emqc_ds *handle, callback func);

/**
 *  @ brief Set generic callback.
 *  @ param handle : an emqc device shadow client instance.
 *  @ param func : a callback function for generic data.
 */
void emqc_ds_set_generic_callback(emqc_ds *handle, callback func);

/**
 *  @ brief Set control callback.
 *  @ param handle : an emqc device shadow client instance.
 *  @ param func : a callback function for control data.
 */
void emqc_ds_set_control_callback(emqc_ds *handle, callback func);

/**
 *  @ brief Update information to shadow.
 *  @ param handle : an emqc device shadow client instance.
 *  @ ret : server response.
 */
EMQC_ERR_CODE emqc_ds_asyn_get(emqc_ds *handle);

/**
 *  @ brief Update information to shadow.
 *  @ param handle : an emqc device shadow client instance.
 *  @ param request : server request.
 *  @ ret : error code.
 */
EMQC_ERR_CODE emqc_ds_asyn_update(emqc_ds *handle, serv_req *resquest);

/**
 *  @ brief Delete shadow data.
 *  @ param handle : an emqc device shadow client instance.
 *  @ param request : server request.
 *  @ ret : error code.
 */
EMQC_ERR_CODE emqc_ds_asyn_delete(emqc_ds *handle, serv_req *request);

/**
 * @ brief new serv_req object.
 * @ ret : serv_req object.
 */
serv_req *new_request();

/**
 *  @ brief Set request current version.
 *  @ param request : server request.
 *  @ param version : current version.
 *  @ ret : void.
 */
void set_request_version(serv_req *request, long version);

/**
 * @ brief Free memory serv_req object.
 * @ param request : serv_req object.
 */
void free_request(serv_req *request);

/**
 *  @ brief Free memory of all attributes.
 *  @ param attr : attributes of array.
 */
void free_attr_vec(attr **attr_);

/**
 *  @ brief Get method from server response
 *  @ param response : server response
 *  @ ret : method
 */
const char *get_param_method(serv_resp *response);


/**
 *  @ brief Get all type and structure with key && value
 *  @ param response : server response
 *  @ param s : desired or reported
 *  @ ret : attribute int array
 */
attr **get_all_param_attr(serv_resp *resp, emqc_ds_state s);

/**
 *  @ brief Get current version from response
 *  @ param re : server response
 *  @ ret : version
 */
long get_version(serv_resp *re);

/**
 *  @ brief Raw payload from response
 *  @ param re : server response
 *  @ ret : raw payload.
 */
char *get_payload(serv_resp *re);

/**
 * @ brief Add null to server request for delete
 * @ param request : server request
 * @ param key : key
 * @ param s : state desired or reported
 * @ ret :
 */
int add_param_null(serv_req *request, const char *key, emqc_ds_state s);

/**
 * @ brief Add int type to server request
 * @ param request : server request
 * @ param key : key
 * @ param val : value
 * @ param s : state desired or reported
 * @ ret :
 */
int add_param_int(serv_req *request, const char *key, int val, emqc_ds_state s);

/**
 * @ brief Add string type to server request
 * @ param request : server request
 * @ param key : key
 * @ param val : value
 * @ param s : state desired or reported
 * @ ret :
 */
int add_param_str(serv_req *request, const char *key, const char *val,
                  emqc_ds_state s);

/**
 * @ brief Add array type to server request
 * @ param request : server request
 * @ param key : key
 * @ param TAG: array type
 * @ param val : value array
 * @ param len : value length
 * @ param s : state desired or reported
 * @ ret :
 */
int add_param_arr(serv_req *request, const char *key, void *val, TAG type,
                  int len, emqc_ds_state s);

#ifdef BUILD_TEST

bool         get_del_rep(serv_req *re_test);
bool         get_del_des(serv_req *re_test);
long         get_version_req(serv_req *re_test);
json_object *get_json_object(serv_req *re_test);
json_object *get_param(serv_req *request, const char *key, emqc_ds_state s);
serv_resp *  new_response(const char *method, int req_id, long timestamp,
                          json_object *jso);

#endif

#endif
