#ifndef EMQC_BASE_H
#define EMQC_BASE_H

#include <MQTTClient.h>

typedef void *(*emqc_cb)(void *data);

typedef enum {
    EMQC_SUCCESS,
    EMQC_CONN_ERR,
    EMQC_DISCONN_ERR,
    EMQC_GET_ERR,
    EMQC_UPDATE_ERR,
    EMQC_DELETE_ERR,
    EMQC_FILED_MISSING_ERR,
    EMQC_INVALID_JSON_ERR,
    EMQC_SUBSCRIBE_FAILED_ERR,
    EMQC_PUBLISH_FAILED_ERR,
    EMQC_MEMORY_ALLOC_FAILED_ERR,
    EMQC_CLIENT_CREATE_ERR,
} EMQC_ERR_CODE;

typedef MQTTClient                Client;
typedef MQTTClient_connectOptions Conn_opts;
volatile MQTTClient_deliveryToken deliveredtoken;

typedef struct {
    char *     product_key;
    char *     device_name;
    int        reconn_interval;
    int        reconn_times;
    Client *   client;
    Conn_opts *conn_opts;
} emqc;

emqc *emqc_alloc(void);

/**
 *  @ brief Set product key.
 *  @ param handle : emqc client instance.
 *  @ param product_key : product key.
 *  @ ret :
 */
void emqc_set_product_key(emqc *handle, char *product_key);

/**
 *  @ brief Set device name.
 *  @ param handle : emqc client instance.
 *  @ param device_name : device name.
 *  @ ret :
 */
void emqc_set_device_name(emqc *handle, char *device_name);

/**
 *  @ brief Set secret.
 *  @ param handle : an emqc client instance.
 *  @ param secert : secret.
 */
void emqc_set_secret(emqc *handle, char *password);

/**
 *  @ brief set reconnect interval and retry times.
 *  @ param handle : an emqc instance.
 *  @ param interval : reconnect retry interval.
 *  @ param handle : reconnect retry times.
 */
// void emqc_ds_set_reconnect(emqc *handle, int interval, int times);

/**
 *  @ brief Initialize emqc client instance.
 *  @ param handle: emqc handle.
 *  @ param host : hostname or ip address of the broker to connect to.
 *  @ param port : network port to connect to.  Usually 1883.
 *  @ ret : error code.
 */
EMQC_ERR_CODE emqc_init(emqc *handle, const char *host, int port);

/**
 *  @ brief Set reconnect interval and retry times.
 *  @ param handle : an emqc client instance.
 *  @ param interval : reconnect retry interval.
 *  @ param handle : reconnect retry times.
 */
void emqc_set_reconnect(emqc *handle, int interval, int times);

/**
 * @ brief Print error code information.
 * @ param error : error code
 */
void emqc_err_code_info(EMQC_ERR_CODE error);

/**
 *  @ brief Connect to an mqtt broker.
 *  @ param handle : an emqc client instance.
 *  @ param keep_alive_interval : keep alive interval.
 *  @ param clean_session : clean session.
 *  @ ret : error code.
 */
EMQC_ERR_CODE emqc_connect(emqc *handle, int keep_alive_interval,
                           int clean_session);

/**
 *  @ brief Subscribe topic.
 *  @ param handle : an emqc client instance.
 *  @ param topic: topic.  
 *  @ ret : err_code.
 */
EMQC_ERR_CODE emqc_sub(emqc *handle, const char *topic);

/**
 *  @ brief Subscribe many topics.
 *  @ param handle : an emqc client instance.
 *  @ param topic: topic array.  
 *  @ param count: topic array count.
 *  @ ret : err_code.
 */
EMQC_ERR_CODE emqc_sub_many(emqc *handle, char *const *topic, int count);

/**
 *  @ brief Publish topic.
 *  @ param handle : an emqc client instance.
 *  @ param topic: topic array.  
 *  @ param payload: payload.
 *  @ ret : err_code.
 */
EMQC_ERR_CODE emqc_pub(emqc *handle, const char *topic, const char *payload);

/**
 *  @ brief Publish many topics.
 *  @ param handle : an emqc client instance.
 *  @ param topic: topic array.  
 *  @ param payload: payload array.
 *  @ param count: topic count.
 *  @ ret : err_code.
 */
EMQC_ERR_CODE emqc_pub_many(emqc *handle, const char **topic,
                            const char **payload, int count);

/**
 *  @ brief Disconnect from a broker.
 *  @ param handle : an emqc client instance.
 *  @ ret : err_code.
 */
EMQC_ERR_CODE emqc_disconnect(emqc *handle);

/**
 *  @ brief Use to free memory associated with a emqc client instance.
 *  @ param handle : emqc client instance.
 *  @ ret :
 */
void emqc_destroy(emqc *handle);

/**
 * @ brief make error code to string
 * @ param EMQC_ERR_CODE : error code
 */
void err_code_info(EMQC_ERR_CODE error);
#endif
