#include <MQTTClient.h>
#include <assert.h>
#include <device_shadow.h>
#include <stdbool.h>
#include <unistd.h>
#include <zmalloc.h>

#define TOPIC_LEN 128

typedef enum {
    DS_GET,
    DS_CONTROL,
    DS_GENERIC,
    DS_UPDATE,
    DS_DELETE
} emqc_ds_method;

struct serv_resp {
    char *       method;
    long         timestamp;
    long         version;
    json_object *payload;
};

struct serv_req {
    bool         del_rep; // default is false
    bool         del_des; // default is false
    long         version;
    json_object *jso_prop;
};

/* Delivered confirmd */
static void delivered(void *context, MQTTClient_deliveryToken dt)
{
    debug("Message with token value %d deliveried", dt);
    deliveredtoken = dt;
}

/* Set timestamp to server response */
static inline int set_param_timestamp(serv_resp *re, long timestamp)
{
    assert(re);
    re->timestamp = timestamp;
    return 0;
}

/* Set payload to server response */
static inline int set_param_json_payload(serv_resp *re, json_object *payload)
{
    assert(re && payload);
    re->payload = payload;
    return 0;
}

/* Set method to server response */
static inline int set_param_method(serv_resp *re, const char *method)
{
    assert(re && method);
    re->method = zstrdup(method);
    return 0;
}

/* Set version to server response */
static inline int set_param_version(serv_resp *re, long version)
{
    assert(re);
    re->version = version;
    return 0;
}

/* Parse json string to struct server response */
static serv_resp *json_parser(char *json)
{
    assert(json);
    json_object *jso           = json_tokener_parse(json);
    json_object *jso_method    = json_object_object_get(jso, "method");
    json_object *jso_payload   = json_object_object_get(jso, "payload");
    json_object *jso_timestamp = json_object_object_get(jso, "timestamp");
    json_object *jso_version   = json_object_object_get(jso, "version");

    serv_resp *resp = (serv_resp *) zmalloc(sizeof(serv_resp));
    if (resp == NULL) {
        log_err("Memory error!");
        return NULL;
    }

    if (jso_method) {
        const char *method = json_object_get_string(jso_method);
        set_param_method(resp, method);
        debug("[method]: %s", resp->method);
    } else {
        log_err("No method!");
    }

    if (jso_payload) {
        json_object *jso_p = json_object_get(jso_payload);
        set_param_json_payload(resp, jso_p);
        debug("[payload]: %s", json_object_to_json_string(resp->payload));
    } else {
        log_err("Payload is empty!");
    }

    if (jso_timestamp) {
        long timestamp = json_object_get_int(jso_timestamp);
        set_param_timestamp(resp, timestamp);
        debug("[timestamp]: %ld", resp->timestamp);
    } else {
        log_err("Timestamp is NULL!");
    }

    if (jso_version) {
        long version = json_object_get_int(jso_version);
        debug("[version]: %ld", version);
        set_param_version(resp, version);
    } else {
        set_param_version(resp, -2);
    }

    json_object_put(jso);

    return resp;
}

/* Free struct server_response */
static void free_response(serv_resp *resp)
{
    if (resp) {
        if (resp->method) {
            zfree(resp->method);
            resp->method = NULL;
        }

        if (resp->payload) {
            json_object_put(resp->payload);
        }

        zfree(resp);
        resp = NULL;
    }
}
/* Different callback functions will be called according to different methods
 * and payloads */
static int message_dispatch(void *context, char *topicName, int topicLen,
                            MQTTClient_message *message)
{
    emqc_ds *m = (emqc_ds *) context;
    debug("message: %.*s", message->payloadlen, (char *) message->payload);
    // log_info("message: %.*s", message->payloadlen, (char*)message->payload);
    char *load = zmalloc(message->payloadlen + 1);
    if (load == NULL) {
        log_err("Memory alloc failed");
        return 1;
    }

    char *msg = (char *) message->payload;
    log_info("#RECV PAYLOAD: %s", msg);

    for (int i = 0; i < message->payloadlen; i++) {
        load[i] = msg[i];
    }

    load[message->payloadlen] = '\0';

    serv_resp *resp = json_parser(load);

    if (resp == NULL) {
        log_err("Response is NULL");
        goto exit;
    }

    if (resp->method && !strcmp(resp->method, "control")) {
        if (m->cb[DS_CONTROL]) {
            m->cb[DS_CONTROL]((void *) resp);
        }
    } else if (resp->method && !strcmp(resp->method, "reply")) {
        if (json_object_object_get_ex(resp->payload, "state", NULL)) {
            if (m->cb[DS_GET]) {
                m->cb[DS_GET]((void *) resp);
            }
        } else {
            if (m->cb[DS_GENERIC]) {
                m->cb[DS_GENERIC]((void *) resp);
            }
        }
    }

exit:

    zfree(load);
    free_response(resp);
    MQTTClient_freeMessage(&message);
    MQTTClient_free(topicName);
    return 1;
}

/* When connection lost, reconnect will be executed according to reconnect
 * parameters */
static void reconnect(void *context, char *cause)
{
    static int time = 0;
    log_warn("Connection lost: %s", cause);

    emqc_ds *ds = (emqc_ds *) context;
    emqc *   m  = ds->handle;

    if (time < m->reconn_times) {
        sleep(m->reconn_interval);
        Conn_opts *opts = m->conn_opts;
        emqc_connect(m, opts->keepAliveInterval, opts->cleansession);
        time++;
    } else {
        log_err("Reconnection failed!");
        time = 0;
    }
}

/**
 *  @ brief create a new emqc device shadow client instance.
 *  @ ret : emqc device shadow client instance.
 */
emqc_ds *emqc_ds_new()
{
    emqc_ds *ds = (emqc_ds *) zmalloc(sizeof(emqc_ds));
    if (ds == NULL) {
        log_err("error memory alloc!");
        return NULL;
    }

    emqc *handle = emqc_alloc();
    if (handle == NULL) {
        log_err("emqc alloc failed!");
        zfree(ds);
        ds = NULL;
    }

    ds->handle    = handle;
    ds->sub_topic = NULL;
    ds->pub_topic = NULL;
    memset(ds->cb, 0, 3 * sizeof(callback));

    return ds;
}

/**
 *  @ brief initialize emqc device shadow client instance.
 *  @ param handle: emqc device shadow client instance.
 *  @ param host : hostname or ip address of the broker to connect to.
 *  @ param port : network port to connect to.  usually 1883.
 *  @ ret : error code.
 */
emqc_ds *emqc_ds_init(emqc_ds *ds, char *host, int port)
{
    assert(ds && host);
    if (EMQC_SUCCESS != emqc_init(ds->handle, host, port)) {
        log_err("emqc init failed!");
        return NULL;
    }

    return ds;
}

/**
 *  @ brief use to free memory associated with a emqc device shadow client
 * instance.
 *  @ param handle : emqc device shadow client instance.
 *  @ ret :
 */
void emqc_ds_destroy(emqc_ds *ds)
{
    if (ds) {
        emqc_destroy(ds->handle);
        if (ds->sub_topic) {
            zfree(ds->sub_topic);
            ds->sub_topic = NULL;
        }

        if (ds->pub_topic) {
            zfree(ds->pub_topic);
            ds->pub_topic = NULL;
        }

        zfree(ds);
        ds = NULL;
    }
}

/**
 *  @ brief set topic.
 *  @ param handle : emqc device shadow client instance.
 *  @ param topic : topic.
 *  @ ret : error code.
 */
static EMQC_ERR_CODE emqc_ds_set_topic(emqc_ds *ds)
{
    assert(ds);
    char  pub_topic[TOPIC_LEN];
    char  sub_topic[TOPIC_LEN];
    emqc *handle = ds->handle;
    if (!handle->device_name || !handle->product_key) {
        log_err("device_name or product_key is missing");
        return EMQC_FILED_MISSING_ERR;
    }
    sprintf(pub_topic, "/fabric/shadow/update/%s/%s", handle->product_key,
            handle->device_name);
    sprintf(sub_topic, "/fabric/shadow/get/%s/%s", handle->product_key,
            handle->device_name);
    debug("pub_topic: %s", pub_topic);
    debug("sub_topic: %s", sub_topic);
    ds->sub_topic = zstrdup(sub_topic);
    ds->pub_topic = zstrdup(pub_topic);
    return EMQC_SUCCESS;
}

/**
 *  @ brief connect to an mqtt broker.
 *  @ param handle : an emqc device shadow client instance.
 *  @ param keep_alive_interval : keep alive interval.
 *  @ param clean_session : clean session.
 *  @ ret : error code.
 */
EMQC_ERR_CODE emqc_ds_connect(emqc_ds *ds, int keep_alive_interval,
                              int clean_session)
{
    int   rc;
    emqc *handle = ds->handle;
    if ((rc = MQTTClient_setCallbacks(*handle->client, ds, reconnect,
                                      message_dispatch, delivered)) !=
        MQTTCLIENT_SUCCESS) {
        log_err("Failed to set callbacks, return code %d\n", rc);
        rc = EMQC_CONN_ERR;
        goto exit;
    }

    Conn_opts *opts = ds->handle->conn_opts;

    if ((rc = emqc_connect(handle, opts->keepAliveInterval,
                           opts->cleansession)) != EMQC_SUCCESS) {
        log_err("Connect failed!");
        goto exit;
    }

    if ((rc = emqc_ds_set_topic(ds)) != EMQC_SUCCESS) {
        goto exit;
    }

    rc = emqc_sub(handle, ds->sub_topic);
    if (rc != MQTTCLIENT_SUCCESS) {
        rc = EMQC_CONN_ERR;
    }

exit:
    return rc;
}

/**
 *  @ brief set get callback.
 *  @ param handle : an emqc device shadow client  instance.
 *  @ param func : a callback function for receive data.
 */
void emqc_ds_set_get_callback(emqc_ds *ds, callback func)
{
    ds->cb[DS_GET] = func;
}

/**
 *  @ brief set generic callback.
 *  @ param handle : an emqc device shadow client instance.
 *  @ param func : a callback function for generic data.
 */
void emqc_ds_set_generic_callback(emqc_ds *ds, callback func)
{
    ds->cb[DS_GENERIC] = func;
}

/**
 *  @ brief set control callback.
 *  @ param handle : an emqc device shadow client instance.
 *  @ param func : a callback function for control data.
 */
void emqc_ds_set_control_callback(emqc_ds *ds, callback func)
{
    ds->cb[DS_CONTROL] = func;
}

/**
 *  @ brief send request infomation to topic
 *  @ param payload: request infomation
 *  @ param topic : publish topic
 *  @ ret :
 */
static int send_raw(emqc_ds *ds, serv_req *request, emqc_ds_method index)
{
    emqc *       handle = ds->handle;
    json_object *jso    = json_object_new_object();

    if (index == DS_GET) {
        json_object_object_add(jso, "method", json_object_new_string("get"));
    } else {

        if (index == DS_UPDATE) {
            json_object_object_add(jso, "method",
                                   json_object_new_string("update"));
        } else {
            json_object_object_add(jso, "method",
                                   json_object_new_string("delete"));
        }

        if (request->del_des) {
            json_object *jso_null = json_object_new_string("null");
            json_object_object_add(request->jso_prop, "desired", jso_null);
        }
        if (request->del_rep) {
            json_object *jso_null = json_object_new_string("null");
            json_object_object_add(request->jso_prop, "reported", jso_null);
        }

        json_object_object_add(jso, "state", request->jso_prop);
        request->jso_prop = NULL;
        json_object_object_add(jso, "version",
                               json_object_new_int(request->version));
    }

    const char *payload = json_object_to_json_string(jso);
    // log_info("send payload: %s", payload);

    int rc = emqc_pub(handle, ds->pub_topic, payload);
    json_object_put(jso);
    return rc;
}

/**
 *  @ brief update information to shadow.
 *  @ param handle : an emqc device shadow client  instance.
 *  @ ret : error code.
 */
EMQC_ERR_CODE emqc_ds_asyn_get(emqc_ds *ds)
{
    assert(ds);

    int rc = send_raw(ds, NULL, DS_GET);
    if (rc != MQTTCLIENT_SUCCESS) {
        rc = EMQC_GET_ERR;
    }

    return rc;
}

EMQC_ERR_CODE emqc_ds_asyn_update(emqc_ds *ds, serv_req *request)
{
    assert(ds && request);

    int rc = send_raw(ds, request, DS_UPDATE);
    if (rc != MQTTCLIENT_SUCCESS) {
        rc = EMQC_UPDATE_ERR;
    }

    return rc;
}

/**
 *  @ brief delete shadow data.
 *  @ param handle : an emqc device shadow client instance.
 *  @ param request : server request.
 *  @ ret : error code.
 */
EMQC_ERR_CODE emqc_ds_asyn_delete(emqc_ds *ds, serv_req *request)
{
    assert(ds && request);

    int rc = send_raw(ds, request, DS_DELETE);

    if (rc != MQTTCLIENT_SUCCESS) {
        rc = EMQC_DELETE_ERR;
    }

    return rc;
}

/* new attr object */
static inline attr *new_attr(const char *key, void *val, TAG tag)
{
    attr *attr_ = (attr *) zmalloc(sizeof(attr));
    attr_->tag  = tag;
    attr_->key  = zstrdup(key);

    if (tag == INT) {
        attr_->val_int = *(int *) val;
    } else if (tag == STR) {
        attr_->val_str = zstrdup((char *) val);
    } else if (tag == ARR) {
        attr_->val_arr = (attr_in *) val;
    }
    attr_->timestamp = 0;

    return attr_;
}

/**
 *  @ brief free memory of all attributes.
 *  @ param attr : attributes of array.
 */
void free_attr_vec(attr **v)
{
    if (v) {
        for (int i = 0; i < cvector_size(v); i++) {
            if (v[i]->key) {
                zfree(v[i]->key);
                v[i]->key = NULL;
            }

            if (v[i]->tag == STR) {
                if (v[i]->val_str) {
                    zfree(v[i]->val_str);
                    v[i]->val_str = NULL;
                }
            } else if (v[i]->tag == ARR) {
                if (v[i]->val_arr->tag == STR) {
                    for (int j = 0; j < cvector_size(v[i]->val_arr); j++) {
                        if (v[i]->val_arr[j].val_str) {
                            zfree(v[i]->val_arr[j].val_str);
                        }
                    }
                }
                cvector_free(v[i]->val_arr);
            }
            zfree(v[i]);
        }

        cvector_free(v);
    }
}

/* get all attributes from json object */
static int json_iter(json_object *jso, attr ***attr_vec)
{
    struct json_object_iterator it    = json_object_iter_begin(jso);
    struct json_object_iterator itend = json_object_iter_end(jso);

    while (!json_object_iter_equal(&it, &itend)) {

        const char * key = json_object_iter_peek_name(&it);
        json_object *val = json_object_iter_peek_value(&it);

        debug("key:%s", key);
        attr *attr_tmp = NULL;
        switch ((json_object_get_type(val))) {
        case json_type_int:;
            int val_int = json_object_get_int(val);
            attr_tmp    = new_attr(key, (void *) &val_int, INT);
            debug("value: %d", attr_tmp->val_int);
            cvector_push_back((*attr_vec), attr_tmp);
            break;

        case json_type_string:;
            const char *val_str = json_object_get_string(val);
            attr_tmp            = new_attr(key, (void *) val_str, STR);
            debug("value: %s", attr_tmp->val_str);
            cvector_push_back((*attr_vec), attr_tmp);
            break;

        case json_type_object:
            json_iter(val, attr_vec);
            break;

        case json_type_array:;
            array_list *al  = json_object_get_array(val);
            int         len = array_list_length(al);
            attr_in *   vec = NULL;

            for (int i = 0; i < len; i++) {
                json_object *j = (json_object *) array_list_get_idx(al, i);
                if (json_type_string == json_object_get_type(j)) {
                    const char *t = json_object_get_string(j);
                    debug("value: %s", t);
                    attr_in attr_in_;
                    attr_in_.tag     = STR;
                    attr_in_.val_str = zstrdup(t);
                    cvector_push_back(vec, attr_in_);
                }

                if (json_type_int == json_object_get_type(j)) {
                    int val_int = json_object_get_int(j);
                    debug("value: %d", val_int);
                    attr_in attr_in_;
                    attr_in_.tag     = INT;
                    attr_in_.val_int = val_int;
                    cvector_push_back(vec, attr_in_);
                }
            }

            attr_tmp = new_attr(key, (void *) vec, ARR);
            cvector_push_back((*attr_vec), attr_tmp);
            break;
        case json_type_double:
            debug("json_type_double");
            // todo
            break;

        default:
            debug("not support!");
            break;
        }

        json_object_iter_next(&it);
    }

    return 0;
}

/* get timestamp by key from server response */
static int json_get_timestamp(json_object *json_meta, attr *attr_vec[])
{
    for (int i = 0; i < cvector_size(attr_vec); i++) {
        json_object *jso;
        if (json_object_object_get_ex(json_meta, attr_vec[i]->key, &jso)) {
            json_object *j;
            if (json_object_object_get_ex(jso, "timestamp", &j)) {
                attr_vec[i]->timestamp = json_object_get_int(j);
            }
        }
    }

    return 0;
}

/**
 *  @ brief get all type and structure with key && value
 *  @ param response : server response
 *  @ param s : desired or reported
 *  @ ret : attribute int array
 */
attr **get_all_param_attr(serv_resp *re, emqc_ds_state s)
{
    assert(re);
    attr **ret = NULL;

    json_object *jso        = NULL;
    json_object *jso_state  = NULL;
    json_object *jso_meta   = NULL;
    json_object *jso_status = NULL;

    // debug("json string : %s", json_object_to_json_string(re->payload));

    if (json_object_object_get_ex(re->payload, "state", &jso_state)) {
        if (s == REPORTED) {
            jso = json_object_object_get(jso_state, "reported");
        } else if (s == DESIRED) {
            jso = json_object_object_get(jso_state, "desired");
        }

        if (jso) {
            json_iter(jso, &ret);
            jso = NULL;
        }
    }

    if (json_object_object_get_ex(re->payload, "metadata", &jso_meta)) {
        if (s == REPORTED) {
            jso = json_object_object_get(jso_meta, "reported");
        } else if (s == DESIRED) {
            jso = json_object_object_get(jso_meta, "desired");
        }

        if (jso) {
            json_get_timestamp(jso, ret);
            jso = NULL;
        }
    }

    if (json_object_object_get_ex(re->payload, "status", &jso_status)) {
        const char *val         = json_object_get_string(jso_status);
        attr *      attr_status = new_attr("status", (void *) val, STR);
        attr_status->timestamp  = re->timestamp;
        cvector_push_back(ret, attr_status);
    }

    return ret;
}

/**
 *  @ brief get current version from response
 *  @ param re : server response
 *  @ ret : version
 */
long get_version(serv_resp *re)
{
    return re->version;
}

/**
 *  @ brief raw payload from response
 *  @ param re : server response
 *  @ ret : raw payload.
 */
char *get_payload(serv_resp *re)
{
    const char *payload = json_object_get_string(re->payload);
    char *      ret     = zstrdup(payload);
    return ret;
}

/**
 * @ brief new serv_req object.
 * @ ret : serv_req object.
 */
serv_req *new_request()
{
    serv_req *   request  = (serv_req *) zmalloc(sizeof(serv_req));
    json_object *jso_prop = json_object_new_object();
    request->jso_prop     = jso_prop;
    request->version      = 0;
    request->del_des      = false;
    request->del_rep      = false;
    return request;
}

/**
 *  @ brief set request current version.
 *  @ param request : server request.
 *  @ param version : current version.
 *  @ ret : void.
 */
void set_request_version(serv_req *request, long version)
{
    assert(request);
    request->version = version;
}

/**
 * @ brief free memory serv_req object.
 * @ param request : serv_req object.
 */
void free_request(serv_req *request)
{
    if (request) {
        if (request->jso_prop) {
            json_object_put(request->jso_prop);
            request->jso_prop = NULL;
        }
        zfree(request);
        request = NULL;
    }
}

/**
 * get json_object attributes ,if it's NULL,
 * add new json_object and return
 */
static json_object *json_get(json_object *prop, emqc_ds_state s)
{
    json_object *rep_or_des = NULL;

    if (s == REPORTED) {
        if (!json_object_object_get_ex(prop, "reported", &rep_or_des)) {
            rep_or_des = json_object_new_object();
            json_object_object_add(prop, "reported", rep_or_des);
        }

    } else {
        if (!json_object_object_get_ex(prop, "desired", &rep_or_des)) {
            rep_or_des = json_object_new_object();
            json_object_object_add(prop, "desired", rep_or_des);
        }
    }

    return rep_or_des;
}

/**
 * @ brief add NULL to server request for delete
 * @ param request : server request
 * @ param key : key
 * @ param s : state desired or reported
 * @ ret :
 */
int add_param_null(serv_req *request, const char *key, emqc_ds_state s)
{
    assert(request);
    if (s == REPORTED && key == NULL) {
        request->del_rep = true;
    } else if (s == DESIRED && key == NULL) {
        request->del_des = true;
    } else {
        json_object *attrs    = json_get(request->jso_prop, s);
        json_object *jso_null = json_object_new_string("null");
        json_object_object_add(attrs, key, jso_null);
    }

    return 0;
}

/**
 * @ brief add int type to server request
 * @ param request : server request
 * @ param key : key
 * @ param val : value
 * @ param s : state desired or reported
 * @ ret :
 */
int add_param_int(serv_req *request, const char *key, int val, emqc_ds_state s)
{
    assert(request && key);
    json_object *attrs = json_get(request->jso_prop, s);
    json_object_object_add(attrs, key, json_object_new_int(val));
    // printf("payload: %s", json_object_to_json_string(request->jso_prop));
    return 0;
}

/**
 * @ brief add string type to server request
 * @ param request : server request
 * @ param key : key
 * @ param val : value
 * @ param s : state desired or reported
 * @ ret :
 */
int add_param_str(serv_req *request, const char *key, const char *val,
                  emqc_ds_state s)
{
    assert(request && key && val);
    json_object *attrs = json_get(request->jso_prop, s);
    json_object_object_add(attrs, key, json_object_new_string(val));
    return 0;
}

/**
 * @ brief add array type to server request
 * @ param request : server request
 * @ param key : key
 * @ param tag: array type
 * @ param val : value array
 * @ param len : value length
 * @ param s : state desired or reported
 * @ ret :
 */
int add_param_arr(serv_req *request, const char *key, void *val, TAG type,
                  int len, emqc_ds_state s)
{
    assert(request && key && val);
    json_object *attrs = json_get(request->jso_prop, s);
    json_object *array = json_object_new_array();

    if (type == INT) {
        int *val_arr = (int *) val;
        for (int i = 0; i < len; i++) {
            json_object_array_add(array, json_object_new_int(val_arr[i]));
        }
    } else if (type == STR) {
        char **val_str = (char **) val;
        for (int i = 0; i < len; i++) {
            json_object_array_add(array, json_object_new_string(val_str[i]));
        }
    } else {
        log_err("non support json value type");
        return EMQC_INVALID_JSON_ERR;
    }

    json_object_object_add(attrs, key, array);
    return EMQC_SUCCESS;
}

#ifdef BUILD_TEST
bool get_del_rep(serv_req *re_test)
{
    return re_test->del_rep;
}

bool get_del_des(serv_req *re_test)
{
    return re_test->del_des;
}

long get_version_req(serv_req *re_test)
{
    return re_test->version;
}

json_object *get_json_object(serv_req *re_test)
{
    return re_test->jso_prop;
}

serv_resp *new_response(const char *method, int version, long timestamp,
                        json_object *jso)
{
    serv_resp *resp = (serv_resp *) zmalloc(sizeof(serv_resp));
    set_param_method(resp, method);
    set_param_version(resp, version);
    set_param_timestamp(resp, timestamp);
    set_param_json_payload(resp, jso);
    return resp;
}

#endif
