#include "emqc_base.h"
#include "dbg.h"
#include <assert.h>
#include <openssl/md5.h>
#include <unistd.h>
#include <zmalloc.h>

emqc *emqc_alloc()
{
    emqc *handle = (emqc *) zmalloc(sizeof(emqc));
    if (handle == NULL) {
        log_err("Error memory alloc!");
        return NULL;
    }

    handle->product_key     = NULL;
    handle->device_name     = NULL;
    handle->client          = NULL;
    handle->reconn_interval = 0;
    handle->reconn_times    = 0;
    handle->conn_opts       = (Conn_opts *) zmalloc(sizeof(Conn_opts));

    if (handle->conn_opts == NULL) {
        log_err("Memory alloc error!");
        return NULL;
    }
    Conn_opts conn_opts = MQTTClient_connectOptions_initializer;
    memcpy(handle->conn_opts, &conn_opts, sizeof(Conn_opts));

    return handle;
}

static char *emqc_str2md5(const char *str, int length)
{
    int           n;
    MD5_CTX       c;
    unsigned char digest[16];
    char *        out = (char *) zmalloc(33);

    MD5_Init(&c);

    while (length > 0) {
        if (length > 512) {
            MD5_Update(&c, str, 512);
        } else {
            MD5_Update(&c, str, length);
        }
        length -= 512;
        str += 512;
    }

    MD5_Final(digest, &c);

    for (n = 0; n < 16; ++n) {
        snprintf(&(out[n * 2]), 16 * 2, "%02x", (unsigned int) digest[n]);
    }

    log_info("Set cliend_id: %s", out);

    return out;
}

/**
 *  @ brief Set product key.
 *  @ param handle : emqc client instance.
 *  @ param product_key : product key.
 *  @ ret :
 */
void emqc_set_product_key(emqc *handle, char *product_key)
{
    assert(handle && product_key);
    handle->product_key = zstrdup(product_key);
}

/**
 *  @ brief Set device name.
 *  @ param handle : emqc client instance.
 *  @ param device_name : device name.
 *  @ ret :
 */
void emqc_set_device_name(emqc *handle, char *device_name)
{
    assert(handle && device_name);
    handle->device_name = zstrdup(device_name);
}

/**
 *  @ brief Set secret.
 *  @ param handle : an emqc client instance.
 *  @ param secert : secret.
 */
void emqc_set_secret(emqc *handle, char *password)
{
    if (!handle && !password) {
        log_err("Handle or password is NULL!");
    }

    handle->conn_opts->password = zstrdup(password);
    log_info("Set secret: %s", handle->conn_opts->password);
}

static void emqc_set_user_name(emqc *handle)
{
    assert(handle && handle->product_key && handle->device_name);

    size_t len = strlen(handle->product_key) + strlen(handle->device_name);
    char * user_name = (char *) zmalloc(len + 2);

    if (user_name == NULL) {
        log_err("Memory alloc error!");
        return;
    }

    snprintf(user_name, len + 2, "%s-%s", handle->product_key,
             handle->device_name);
    handle->conn_opts->username = user_name;
    log_info("Set user_name: %s", handle->conn_opts->username);
}
static char *emqc_get_clien_id(emqc *handle)
{
    assert(handle && handle->conn_opts && handle->conn_opts->username);
    const char *user_name = handle->conn_opts->username;
    return emqc_str2md5(user_name, strlen(user_name));
}

/**
 *  @ brief Initialize emqc client instance.
 *  @ param handle: emqc handle.
 *  @ param host : hostname or ip address of the broker to connect to.
 *  @ param port : network port to connect to.  Usually 1883.
 *  @ ret : error code.
 */
EMQC_ERR_CODE emqc_init(emqc *handle, const char *host, int port)
{
    assert(handle && host);

    int rc = EMQC_SUCCESS;
    emqc_set_user_name(handle);
    char *client_id = emqc_get_clien_id(handle);

    handle->client = (Client *) zmalloc(sizeof(Client));
    if (handle->client == NULL) {
        rc = EMQC_MEMORY_ALLOC_FAILED_ERR;
        goto exit;
    }

    char addr[100];
    sprintf(addr, "tcp://%s:%d", host, port);
    log_info("Connect to tcp://%s:%d", host, port);

    if ((rc = MQTTClient_create(handle->client, addr, client_id,
                                MQTTCLIENT_PERSISTENCE_NONE, NULL)) !=
        MQTTCLIENT_SUCCESS) {
        log_err("Failed to create client, return code %d\n", rc);
        rc = EMQC_CLIENT_CREATE_ERR;
        goto exit;
    }

    zfree(client_id);
    client_id = NULL;

    return EMQC_SUCCESS;

exit:
    if (handle) {
        if (handle->client) {
            zfree(handle->client);
        }

        zfree(handle);
    }
    log_err("New emqc error!");
    // TODO
    return EMQC_SUCCESS;
}

/**
 *  @ brief Set reconnect interval and retry times.
 *  @ param handle : an emqc client instance.
 *  @ param interval : reconnect retry interval.
 *  @ param handle : reconnect retry times.
 */
void emqc_set_reconnect(emqc *handle, int interval, int times)
{
    handle->reconn_interval = interval;
    handle->reconn_times    = times;
}

/**
 *  @ brief Connect to an mqtt broker.
 *  @ param handle : an emqc client instance.
 *  @ param keep_alive_interval : keep alive interval.
 *  @ param clean_session : clean session.
 *  @ ret : error code.
 */
EMQC_ERR_CODE emqc_connect(emqc *handle, int keep_alive_interval,
                           int clean_session)
{
    int rc;
    // Conn_opts conn_opts = MQTTClient_connectOptions_initializer;
    Conn_opts *conn_opts         = handle->conn_opts;
    conn_opts->cleansession      = clean_session;
    conn_opts->keepAliveInterval = keep_alive_interval;

    if ((rc = MQTTClient_connect(*handle->client, conn_opts)) !=
        MQTTCLIENT_SUCCESS) {
        log_err("Failed to connect, return code %d\n", rc);
        log_err("String %s", MQTTReasonCode_toString(rc));
        rc = EMQC_CONN_ERR;
        goto exit;
    }

exit:
    return rc;
}

EMQC_ERR_CODE emqc_sub(emqc *handle, const char *topic)
{
    int rc;
    int qos = 0;
    log_info("topic: %s", topic);

    if ((rc = MQTTClient_subscribe(*handle->client, topic, qos)) !=
        MQTTCLIENT_SUCCESS) {
        log_err("Failed to subscribe, return code %d", rc);
        return EMQC_SUBSCRIBE_FAILED_ERR;
    }

    return EMQC_SUCCESS;
}

EMQC_ERR_CODE emqc_sub_many(emqc *handle, char *const *topic, int count)
{
    int rc;
    if (count <= 0) {
        return EMQC_SUBSCRIBE_FAILED_ERR;
    }

    int qos[count];
    for (int i = 0; i < count; i++) {
        qos[i] = 0;
    }
    // memset(qos, 1, count * sizeof(int));

    if ((rc = MQTTClient_subscribeMany(*handle->client, count, topic, qos)) !=
        MQTTCLIENT_SUCCESS) {
        log_err("Failed to subscribe, return code %d", rc);
        return EMQC_SUBSCRIBE_FAILED_ERR;
    }

    return EMQC_SUCCESS;
}

EMQC_ERR_CODE emqc_pub(emqc *handle, const char *topic, const char *payload)
{
    MQTTClient_message pubmsg = MQTTClient_message_initializer;
    debug("payload: %s\n", payload);
    // log_info("payload: %s\n", payload);
    pubmsg.payload    = (void *) payload;
    pubmsg.payloadlen = (int) strlen(payload);
    pubmsg.qos        = 0;
    pubmsg.retained   = 0;
    MQTTClient_deliveryToken token;

    int rc;

    if ((rc = MQTTClient_publishMessage(*handle->client, topic, &pubmsg,
                                        &token)) != MQTTCLIENT_SUCCESS) {
        log_err("Failed to publish message, return code %d", rc);
    }

    rc = MQTTClient_waitForCompletion(*handle->client, token, 500L);

    return EMQC_SUCCESS;
}

EMQC_ERR_CODE emqc_pub_many(emqc *handle, const char **topic,
                            const char **payload, int count)
{

    int rc = EMQC_SUCCESS;
    for (size_t i = 0; i < count; i++) {
        if ((rc = emqc_pub(handle, *payload++, *topic++)) != EMQC_SUCCESS) {
            return rc;
        }
    }

    return EMQC_SUCCESS;
}

/**
 *  @ brief Disconnect from a broker.
 *  @ param handle : an emqc client instance.
 *  @ ret : err_code.
 */
EMQC_ERR_CODE emqc_disconnect(emqc *handle)
{
    int rc;
    if ((rc = MQTTClient_disconnect(*handle->client, 10000)) !=
        MQTTCLIENT_SUCCESS) {
        log_err("Failed to disconnect, return code %d\n", rc);
        rc = EMQC_DISCONN_ERR;
    }

    return rc;
}

/**
 *  @ brief Use to free memory associated with a emqc client instance.
 *  @ param handle : emqc client instance.
 *  @ ret :
 */
void emqc_destroy(emqc *handle)
{
    if (handle) {
        if (handle->product_key) {
            zfree(handle->product_key);
            handle->product_key = NULL;
        }

        if (handle->device_name) {
            zfree(handle->device_name);
            handle->device_name = NULL;
        }

        if (handle->client) {
            MQTTClient_destroy(handle->client);
            zfree(handle->client);
        }

        if (handle->conn_opts) {
            Conn_opts *conn_opts = handle->conn_opts;

            if (conn_opts->username) {
                zfree((void *) conn_opts->username);
                conn_opts->username = NULL;
            }

            if (conn_opts->password) {
                zfree((void *) conn_opts->password);
                conn_opts->password = NULL;
            }

            zfree(conn_opts);
            conn_opts = NULL;
        }

        zfree(handle);
        handle = NULL;
    }
}


/**
 * @ brief make error code to string
 * @ param EMQC_ERR_CODE : error code
 */
void err_code_info(EMQC_ERR_CODE error)
{
    switch (error) {
        case EMQC_SUCCESS:
            puts("success");
            break;
        case EMQC_CONN_ERR:
            puts("connection error");
            break;
        case EMQC_DISCONN_ERR:
            puts("disconnection error");
            break;
        case EMQC_GET_ERR:
            puts("get error");
            break;
        case EMQC_UPDATE_ERR:
            puts("update error");
            break;
        case EMQC_DELETE_ERR:
            puts("delete error");
            break;
        case EMQC_FILED_MISSING_ERR:
            puts("device_name or product_key is missing");
            break;
        case EMQC_INVALID_JSON_ERR:
            puts("invalid json data!");
            break;
        case EMQC_SUBSCRIBE_FAILED_ERR:
            puts("subscribe failed!");
            break;
        case EMQC_PUBLISH_FAILED_ERR:
            puts("publish failed!");
            break;
        case EMQC_MEMORY_ALLOC_FAILED_ERR:
            puts("memory alloc failed!");
            break;
        case EMQC_CLIENT_CREATE_ERR:
            puts("client create failed!");
            break;
        default:
            puts("unknown error");
    }
}
