// #define _POSIX_C_SOURCE 200809L

#include <assert.h>
#include <cvector.h>
#include <pthread.h>
#include <stdatomic.h>
#include <thing_model.h>
#include <unistd.h>
#include <zmalloc.h>

#include <math.h>
#include <time.h>

struct tm_ctxt {
    char *  sfx;
    emqc_cb cb;
};

typedef struct {
    char  topic[128];
    char *payload;
} pub_info_t;

typedef struct {
    pub_info_t **   que;
    pthread_t       thr;
    pthread_mutex_t mtx;
    pthread_cond_t  cv;
} tm_pub_info_que;

static tm_pub_info_que *tmq  = NULL;
static atomic_int       acnt = 0;

static void emqc_tm_add_param(json_object *param, char *key, emqc_tm_object *obj);
static void emqc_tm_add_struct(json_object *param, const char *key, void *value);

static void *emqc_tm_worker(void *args)
{
    emqc_tm *tm = (emqc_tm *) args;

    for (;;) {
        pub_info_t *t = NULL;
        pthread_mutex_lock(&tmq->mtx);

        size_t size = 0;
        while (((size = cvector_size(tmq->que)) == 0)) {
            log_info("Waitting for task......");
            pthread_cond_wait(&tmq->cv, &tmq->mtx);
        }

        t = tmq->que[size - 1];
        cvector_pop_back(tmq->que);
        pthread_mutex_unlock(&tmq->mtx);

        log_info("#PUB: %s", t->topic);
        // log_info("####: SIZE [%ld], %s", size, payload);
        emqc_tm_pub(tm, t->topic, t->payload);
        if (t->payload) {
            zfree(t->payload);
        }

        zfree(t);
        t = NULL;
    }

    return NULL;
}

emqc_tm *emqc_tm_new(void)
{
    emqc_tm *tm = (emqc_tm *) zmalloc(sizeof(emqc_tm));
    if (!tm) {
        return NULL;
    }

    tm->handle = emqc_alloc();
    if (tm->handle == NULL) {
        goto exit;
    }

    tm->ctxts = NULL;

    return tm;

exit:
    if (tm) {
        emqc_destroy(tm->handle);
        zfree(tm);
        tm = NULL;
    }

    return NULL;
}

emqc_tm *emqc_tm_init(emqc_tm *tm, const char *host, int port)
{
    if (EMQC_SUCCESS != emqc_init(tm->handle, host, port)) {
        log_err("emqc init failed!");
        return NULL;
    }

    tmq      = (tm_pub_info_que *) zmalloc(sizeof(tm_pub_info_que));
    tmq->que = NULL;
    pthread_mutex_init(&(tmq->mtx), NULL);
    pthread_cond_init(&(tmq->cv), NULL);

    if (pthread_create(&(tmq->thr), NULL, emqc_tm_worker, (void *) tm)) {
        log_err("Pthread create failed!");
        goto exit;
    }

    if (pthread_detach(tmq->thr)) {
        log_err("pthread detach: %s", strerror(errno));
        goto exit;
    }

    return tm;
exit:
    emqc_destroy(tm->handle);
    return NULL;
}

static tm_ctxt *emqc_tm_ctxt_new(const char *sfx, emqc_cb cb)
{
    tm_ctxt *ctxt = (tm_ctxt *) zmalloc(sizeof(tm_ctxt));
    if (ctxt == NULL) {
        log_err("Memory alloc error!");
        return NULL;
    }

    ctxt->sfx = zstrdup(sfx);
    ctxt->cb  = cb;

    return ctxt;
}

static void emqc_tm_ctxt_destroy(tm_ctxt *ctxt)
{
    if (ctxt) {
        if (ctxt->sfx) {
            zfree(ctxt->sfx);
            ctxt->sfx = NULL;
        }

        zfree(ctxt);
        ctxt = NULL;
    }
}

void emqc_tm_set_dft_cb(emqc_tm *tm, emqc_tm_cb_opt opt, emqc_cb cb)
{
    if (opt == TM_CB_EVENT_POST_REPLY || opt == TM_CB_ASYN_ACTION_EXEC) {
        log_err("This should be set customize callback!");
        return;
    }
    char *dft_sfx[] = {
        "property/post_reply",
        "property/set",
        "event/post_reply",
        "action/exec",
        "rrpc/request",
        "property/desired/get_reply",
        "property/desired/delete_reply",
    };

    tm_ctxt *ctxt = emqc_tm_ctxt_new(dft_sfx[opt], cb);
    cvector_push_back(tm->ctxts, ctxt);
}

void emqc_tm_set_ctm_cb(emqc_tm *tm, emqc_tm_cb_opt opt, emqc_cb cb,
        char *model_id, char *identifier)
{
    if (opt < TM_CB_EVENT_POST_REPLY || opt > TM_CB_ASYN_ACTION_EXEC) {
        log_err("Non support callback %d for customize!", opt);
        return;
    }

    char *prefix[] = {
        "event",
        "action",
    };

    char *dft_sfx[] = {
        "post_reply",
        "exec",
    };

    char *fmt         = (model_id && identifier) ? "%s/%s:%s/%s" : "%s/%s%s/%s";
    char *mid         = model_id ? model_id : "";
    char *iden        = identifier ? identifier : "";
    char  sfx_buf[64] = { 0 };

    snprintf(sfx_buf, sizeof(sfx_buf), fmt, prefix[opt - 2], mid, iden,
            dft_sfx[opt - 2]);
    tm_ctxt *ctxt = emqc_tm_ctxt_new(sfx_buf, cb);
    cvector_push_back(tm->ctxts, ctxt);
}

static char **emqc_tm_get_sub_topic(emqc_tm *tm, size_t *cnt)
{
    assert(tm);

    emqc *    handle      = tm->handle;
    char *    product_key = handle->product_key;
    char *    device_name = handle->device_name;
    tm_ctxt **ctxts       = tm->ctxts;
    size_t    size        = cvector_size(ctxts);
    *cnt                  = size;

    char **ret = (char **) zmalloc(sizeof(char *) * size);
    if (ret == NULL) {
        return NULL;
    }

    char topic[EMQC_TM_TOPIC_LEN] = { 0 };

    for (int i = 0; i < size; i++) {
        char *sfx = ctxts[i]->sfx;
        if (strstr(sfx, "rrpc")) {
            char request[16];
            snprintf(request, 16, "%s/%s", sfx, "+");
            snprintf(topic, EMQC_TM_TOPIC_LEN, "/fabric/sys/%s/%s/%s",
                    product_key, device_name, request);
        } else {
            snprintf(topic, EMQC_TM_TOPIC_LEN, TM_FMT, product_key, device_name,
                    sfx);
        }
        ret[i] = zstrdup(topic);
        log_info("#SUB: %s", ret[i]);
        memset(topic, 0, EMQC_TM_TOPIC_LEN);
        sfx = NULL;
    }

    return ret;
}

/* Delivered confirmd */
static void delivered(void *context, MQTTClient_deliveryToken dt)
{
    // debug("Message with token value %d deliveried", dt);
    deliveredtoken = dt;
}

static tm_reply *tm_reply_new(const char *id, int code, const char *data,
        const char *msg)
{
    tm_reply *tr = (tm_reply *) zmalloc(sizeof(tm_reply));
    if (tr == NULL) {
        log_err("Memory alloc error!");
        return NULL;
    }
    tr->id   = zstrdup(id);
    tr->code = code;
    tr->data = zstrdup(data);
    if (msg) {
        tr->msg = zstrdup(msg);
    } else {
        tr->msg = NULL;
    }

    return tr;
}

static void tm_reply_destroy(tm_reply *tr)
{
    if (tr) {
        if (tr->id) {
            zfree(tr->id);
            tr->id = NULL;
        }
        if (tr->data) {
            zfree(tr->data);
            tr->data = NULL;
        }
        if (tr->msg) {
            zfree(tr->msg);
            tr->msg = NULL;
        }

        zfree(tr);
        tr = NULL;
    }
}

static tm_request *tm_request_new(const char *id, const char *version,
        const char *method, const char *params)
{
    tm_request *tr = (tm_request *) zmalloc(sizeof(tm_request));
    if (tr == NULL) {
        log_err("Memory alloc error!");
        return NULL;
    }
    tr->id      = zstrdup(id);
    tr->version = zstrdup(version);
    tr->method  = zstrdup(method);
    tr->params  = zstrdup(params);

    return tr;
}

static void tm_request_destroy(tm_request *tr)
{
    if (tr) {
        if (tr->id) {
            zfree(tr->id);
            tr->id = NULL;
        }
        if (tr->version) {
            zfree(tr->version);
            tr->version = NULL;
        }
        if (tr->method) {
            zfree(tr->method);
            tr->method = NULL;
        }
        if (tr->params) {
            zfree(tr->params);
            tr->params = NULL;
        }
        zfree(tr);
        tr = NULL;
    }
}

static emqc_tm_trans_object *emqc_tm_trans_object_new(void *data, int type)
{
    emqc_tm_trans_object *tm = (emqc_tm_trans_object *) zmalloc(sizeof(emqc_tm_trans_object));
    if (tm == NULL) {
        log_err("Memory alloc error!");
        return NULL;
    }

    switch (type) {
        case TM_REP:
            tm->type = TM_REP;
            tm->rep = data;
            break;
        case TM_REQ:
            tm->type = TM_REQ;
            tm->req = data;
            break;
        case TM_RAW:
            tm->type = TM_RAW;
            tm->raw = data;
            break;
        default:
            log_err("Error trans type.");
    }

    return tm;

}

static void emqc_tm_trans_object_destroy(emqc_tm_trans_object *tm)
{
    if (tm) {
        switch (tm->type) {
            case TM_REP:
                tm_reply_destroy(tm->rep);
                break;
            case TM_REQ:
                tm_request_destroy(tm->req);
                break;
            case TM_RAW:
                zfree(tm->raw);
                break;
            default:
                log_err("Error trans type.");
        }
        zfree(tm);
    }
}

static emqc_tm_trans_object *emqc_tm_json_token_parse(char *load)
{
    emqc_tm_trans_object *ret;
    json_object *jso = json_tokener_parse(load);
    if (jso == NULL) {
        char *raw = zstrdup(load);
        ret = emqc_tm_trans_object_new(raw, TM_RAW);
        return ret;
    }

    const char * id  = NULL;
    const char * msg = NULL;
    json_object *jso_id;
    json_object *jso_code;
    json_object *jso_data;
    json_object *jso_msg;
    json_object *jso_version;
    json_object *jso_method;
    json_object *jso_params;

    if (json_object_object_get_ex(jso, "id", &jso_id)) {
        id = json_object_get_string(jso_id);
    } else {
        goto exit;
    }

    if (json_object_object_get_ex(jso, "code", &jso_code)) {
        if (!json_object_object_get_ex(jso, "data", &jso_data)) {
            // log_err("Find code, but no data field!");
            goto exit;
        }

        int code = json_object_get_int(jso_code);
        if (code != 0) {
            if (!json_object_object_get_ex(jso, "message", &jso_msg)) {
                // log_err("Find code, but no data field!");
                goto exit;
            }
            msg = json_object_get_string(jso_msg);
        }

        const char *data = json_object_to_json_string(jso_data);
        // const char *data = json_object_get_string(jso_data);

        void *rep = (void *) tm_reply_new(id, code, data, msg);
        ret = emqc_tm_trans_object_new(rep, TM_REP);
        json_object_put(jso);
        return ret;
    }

    if (json_object_object_get_ex(jso, "params", &jso_params)) {
        if (!json_object_object_get_ex(jso, "method", &jso_method)) {
            goto exit;
        }

        if (!json_object_object_get_ex(jso, "version", &jso_version)) {
            log_err("Find params, but no version field!");
            goto exit;
        }

        const char *version = json_object_get_string(jso_version);
        const char *method  = json_object_get_string(jso_method);
        const char *params  = json_object_to_json_string(jso_params);

        void *req = (void *) tm_request_new(id, version, method, params);
        ret = emqc_tm_trans_object_new(req, TM_REQ);
        json_object_put(jso);
        return ret;
    }

exit:;
    char *raw = zstrdup(load);
    ret = emqc_tm_trans_object_new(raw, TM_RAW);
    json_object_put(jso);
    return ret;
}

static int message_dispatch(void *context, char *topicName, int topicLen,
        MQTTClient_message *message)
{
    emqc_tm *m = (emqc_tm *) context;
    log_info("#RECV FROM: %s", topicName);
    log_info("message: %.*s", message->payloadlen, (char*)message->payload);
    char *load = zmalloc(message->payloadlen + 1);
    if (load == NULL) {
        goto exit;
    }

    char *msg = (char *) message->payload;

    for (int i = 0; i < message->payloadlen; i++) {
        load[i] = msg[i];
    }

    load[message->payloadlen] = '\0';
    // log_info("RECV PAYLOAD: %s", load);

    for (int i = 0; i < cvector_size(m->ctxts); i++) {
        // log_info("%s, %s", topicName, m->ctxts[i]->sfx);
        if (strstr(topicName, m->ctxts[i]->sfx)) {
            emqc_tm_trans_object *payload = emqc_tm_json_token_parse(load);
            if (payload == NULL) {
                log_info("Something wrong!");
                return -1;
            }

            void *ret = m->ctxts[i]->cb(payload);
            if (ret) {
                pub_info_t *pub_info = zmalloc(sizeof(pub_info_t));
                if (pub_info == NULL) {
                    log_err("Memory alloc failed!");
                }

                char *p = NULL;
                if ((p = strstr(topicName, "request")) != NULL) {
                    char prefix[56] = { 0 };
                    char suffix[56] = { 0 };
                    int  len        = p - topicName;

                    snprintf(prefix, len, "%s", topicName);
                    sscanf(p, "request/%s", suffix);

                    snprintf(pub_info->topic, 128, "%s/response/%s", prefix,
                            suffix);
                    puts(pub_info->topic);
                } else {
                    snprintf(pub_info->topic, 128, "%s_reply", topicName);
                }
                pub_info->payload = ret;

                pthread_mutex_lock(&tmq->mtx);
                bool empty = cvector_empty(tmq->que);
                log_info("append task");
                cvector_push_back(tmq->que, pub_info);
                pthread_mutex_unlock(&tmq->mtx);
                if (empty) {
                    log_info("send signal");
                    pthread_cond_signal(&tmq->cv);
                }
            }

            emqc_tm_trans_object_destroy(payload);

            break;
        }
    }

exit:

    zfree(load);
    MQTTClient_freeMessage(&message);
    MQTTClient_free(topicName);
    return 1;
}

/* When connection lost, reconnect will be executed according to reconnect
 * parameters */
static void reconnect(void *context, char *cause)
{
    static int time = 0;
    log_warn("Connection lost: %s", cause);

    emqc_tm *tm = (emqc_tm *) context;
    emqc *   m  = tm->handle;

    if (time < m->reconn_times) {
        sleep(m->reconn_interval);
        Conn_opts *opts = m->conn_opts;
        emqc_connect(m, opts->keepAliveInterval, opts->cleansession);
        time++;
    } else {
        log_err("Reconnection failed!");
        time = 0;
    }
}

EMQC_ERR_CODE emqc_tm_connect(emqc_tm *tm)
{
    int   rc;
    emqc *handle = tm->handle;
    if ((rc = MQTTClient_setCallbacks(*handle->client, tm, reconnect,
                    message_dispatch, delivered)) !=
            MQTTCLIENT_SUCCESS) {
        log_err("Failed to set callbacks, return code %d\n", rc);
        rc = EMQC_CONN_ERR;
        goto exit;
    }

    Conn_opts *opts = tm->handle->conn_opts;

    if ((rc = emqc_connect(handle, opts->keepAliveInterval,
                    opts->cleansession)) != EMQC_SUCCESS) {
        log_err("Connect failed!");
        goto exit;
    }

exit:
    return rc;
}

EMQC_ERR_CODE emqc_tm_sub(emqc_tm *tm)
{
    assert(tm);
    if (tm->handle == NULL) {
        log_err("handle is NULL");
        abort();
    }

    size_t cnt   = 0;
    char **topic = emqc_tm_get_sub_topic(tm, &cnt);

    if (EMQC_SUCCESS != emqc_sub_many(tm->handle, (char *const *) topic, cnt)) {
        log_err("subscribe failed!");
        return EMQC_SUBSCRIBE_FAILED_ERR;
    }

    if (topic != NULL) {
        for (size_t i = 0; i < cnt; i++) {
            if (topic[i] != NULL) {
                zfree(topic[i]);
                topic[i] = NULL;
            }
        }

        zfree(topic);
        topic = NULL;
    }

    return EMQC_SUCCESS;
}

EMQC_ERR_CODE emqc_tm_pub(emqc_tm *tm, const char *topic, const char *payload)
{
    assert(tm && topic);
    if (tm->handle == NULL) {
        log_err("handle is NULL");
        abort();
    }

    if (EMQC_SUCCESS != emqc_pub(tm->handle, topic, payload)) {
        log_err("publish failed!");
        return EMQC_PUBLISH_FAILED_ERR;
    }

    return EMQC_SUCCESS;
}

void emqc_tm_destroy(emqc_tm *tm)
{
    if (tm) {
        emqc_destroy(tm->handle);
        if (tm->ctxts) {
            size_t size = cvector_size(tm->ctxts);
            for (size_t i = 0; i < size; i++) {
                emqc_tm_ctxt_destroy(tm->ctxts[i]);
            }
            cvector_free(tm->ctxts);
        }
        zfree(tm);
        tm = NULL;
    }

    if (tmq) {
        if (tmq->que) {
            for (size_t i = 0; i < cvector_size(tmq->que); i++) {
                zfree(tmq->que[i]);
            }
            cvector_free(tmq->que);
        }
        zfree(tmq);
    }
}

static int emqc_tm_get_id()
{
    return acnt++;
}

static void emqc_tm_concat_method(char *method, size_t n, const char *mtd_dft,
        const char *func_id)
{
    // get method.
    char *fmt = (func_id != NULL) ? "%s.%s" : "%s%s";
    snprintf(method, n, fmt, mtd_dft, func_id ? func_id : "");
}

json_object *emqc_tm_new_payload(const char *mtd_dft, const char *func_id)
{
    char id[8];
    char method[64];

    emqc_tm_concat_method(method, sizeof(method), mtd_dft, func_id);

    // get id.
    snprintf(id, sizeof(id), "%u", emqc_tm_get_id());

    json_object *jso = json_object_new_object();

    json_object_object_add(jso, "id", json_object_new_string(id));
    json_object_object_add(jso, "version", json_object_new_string("1.0"));
    json_object_object_add(jso, "method", json_object_new_string(method));

    return jso;
}

static unsigned long emqc_tm_get_curr_ts_with_ms(void)
{
    long            ms; // Milliseconds
    time_t          s;  // Seconds
    struct timespec spec;

    clock_gettime(CLOCK_REALTIME, &spec);

    s  = spec.tv_sec;
    ms = round(spec.tv_nsec / 1.0e6); // Convert nanoseconds to milliseconds
    if (ms > 999) {
        s++;
        ms = 0;
    }

    unsigned long ts = s * 1000 + ms;

    return ts;
}

// Update timestamp
static void emqc_tm_update_timestamp(json_object *jso)
{
    json_object *ts = NULL;
    // long ts_val = (unsigned long)time(NULL);
    unsigned long ts_val = emqc_tm_get_curr_ts_with_ms();
    // TODO
    // log_info("timestamp: %ld", ts_val);

    if (json_object_object_get_ex(jso, "time", &ts)) {
        json_object_set_int64(ts, ts_val);
    } else {
        json_object_object_add(jso, "time", json_object_new_int64(ts_val));
    }

    return;
}

json_object *emqc_tm_add_get_desired_param(json_object *jso, const char **keys,
        int num_keys, char *func_id)
{
    assert(jso && keys && num_keys > 0);

    json_object *params = json_object_new_array();
    json_object_object_add(jso, "params", params);

    // TODO if strlen(key) + strlen(func_id) > 60;
    if (func_id != NULL) {
        for (int i = 0; i < num_keys; i++) {
            char func_id_key[64];
            snprintf(func_id_key, sizeof(func_id_key), "%s:%s", func_id,
                    keys[i]);
            json_object_array_add(params, json_object_new_string(func_id_key));
            memset(func_id_key, 0, sizeof(func_id_key));
        }

    } else {
        for (int i = 0; i < num_keys; i++) {
            json_object_array_add(params, json_object_new_string(keys[i]));
        }
    }

    return jso;
}

json_object *emqc_tm_add_del_desired_param(json_object *  jso,
        del_des_param *param_arr,
        int num_params, char *func_id)
{
    assert(jso && param_arr && num_params > 0);

    json_object *params = json_object_new_object();
    json_object_object_add(jso, "params", params);

    // TODO if strlen(key) + strlen(func_id) > 60;
    if (func_id != NULL) {
        for (int i = 0; i < num_params; i++) {
            char func_id_key[64];
            snprintf(func_id_key, sizeof(func_id_key), "%s:%s", func_id,
                    param_arr[i].key);
            json_object *param = json_object_new_object();
            if (param_arr[i].version != NULL) {
                json_object_object_add(
                        param, "version",
                        json_object_new_string(param_arr[i].version));
            }

            json_object_object_add(params, func_id_key, param);
            memset(func_id_key, 0, sizeof(func_id_key));
        }

    } else {
        for (int i = 0; i < num_params; i++) {
            json_object *param = json_object_new_object();
            if (param_arr[i].version != NULL) {
                json_object_object_add(
                        param, "version",
                        json_object_new_string(param_arr[i].version));
            }

            json_object_object_add(params, param_arr[i].key, param);
        }
    }

    return jso;
}

static void emqc_tm_add_array(json_object *param, const char *key, void *value, emqc_tm_type etp)
{
    json_object *jso = NULL;
    int i = 0;

    if (!json_object_object_get_ex(param, key, &jso)) {
        jso = json_object_new_array();
        json_object_object_add(param, key, jso);
    }

    switch(etp) {
        case EMQC_TM_ARR_INT:;
            int **param_int = (int **) value;
            do {
                if (param_int[i] == NULL) {
                    break;
                }
                json_object_array_add(jso, json_object_new_int(*(int*)param_int[i]));

            } while (++i);
            break;

        case EMQC_TM_ARR_STR:;
            char **param_str = (char **) value;
            do {
                if (param_str[i] == NULL) {
                    break;
                }
                json_object_array_add(jso, json_object_new_string((char*)param_str[i]));

            } while (++i);
            break;

        case EMQC_TM_ARR_DOUBLE:;
            double **param_dou = (double **) value;
            do {
                if (param_dou[i] == NULL) {
                    break;
                }
                json_object_array_add(jso, json_object_new_double(*(double*)param_dou[i]));

            } while (++i);
            break;

        case EMQC_TM_ARR_STRUCT:;
            emqc_tm_object **tm_object = (emqc_tm_object **) value;
            do {
                if (tm_object[i] == NULL) {
                    break;
                }
                json_object *jso_struct = json_object_new_object();
                emqc_tm_add_struct(jso_struct, NULL, tm_object[i]);
                json_object_array_add(jso, jso_struct);

            } while (++i);
            break;

        default:
            log_err("Not support type!");
    }
}

static void emqc_tm_add_struct(json_object *param, const char *key, void *value)
{
    json_object *jso = NULL;
    int i = 0;

    if (key == NULL) { 
        // for ARR_STRUCT
        jso = param;
    } else {
        // for STRUCT
        if (!json_object_object_get_ex(param, key, &jso)) {
            jso = json_object_new_object();
            json_object_object_add(param, key, jso);
        }
    }
    
    emqc_tm_object *tm_object = (emqc_tm_object *) value;
    do {
        if (tm_object[i].type == EMQC_TM_TERMINATOR) {
            break;
        }
    
        emqc_tm_add_param(jso, tm_object[i].key, &tm_object[i]);
    } while (++i);

}

void emqc_tm_add_service_param(json_object *param, char *key, emqc_tm_object *obj)
{
    emqc_tm_add_param(param, key, obj);
}

void emqc_tm_add_param(json_object *param, char *key, emqc_tm_object *obj)
{
    assert(param && key && obj);

    switch (obj->type) {
        case EMQC_TM_INT:
            json_object_object_add(param, key, json_object_new_int(obj->val_i));
            break;
        case EMQC_TM_STR:
            json_object_object_add(param, key,
                    json_object_new_string(obj->val_s));
            break;
        case EMQC_TM_DOUBLE:
            json_object_object_add(param, key,
                    json_object_new_double(obj->val_d));
            break;

        case EMQC_TM_STRUCT:
            emqc_tm_add_struct(param, key, obj->val_o);
            break;

        case EMQC_TM_ARR_INT:
        case EMQC_TM_ARR_STR:
        case EMQC_TM_ARR_DOUBLE:
        case EMQC_TM_ARR_STRUCT:
            emqc_tm_add_array(param, key, obj->val_o, obj->type);
            break;

        default:
            log_err("Not support type!");
    }
}

json_object *emqc_tm_add_post_param(json_object *jso, emqc_tm_object *obj, char *func_id)
{
    assert(jso && obj);

    json_object *params = NULL;
    // check if params is exist
    if (!json_object_object_get_ex(jso, "params", &params)) {
        params = json_object_new_object();
        json_object_object_add(jso, "params", params);
    }

    json_object *param = NULL;
    char         func_id_key[64];

    if (func_id != NULL) {
        snprintf(func_id_key, sizeof(func_id_key), "%s:%s", func_id, obj->key);
    } else {
        strncpy(func_id_key, obj->key, sizeof(func_id_key));
    }

    // check if key:func_id is not exist, insert.
    if (!json_object_object_get_ex(params, func_id_key, &param)) {
        param = json_object_new_object();
        emqc_tm_add_param(param, "value", obj);
        json_object_object_add(params, func_id_key, param);
    }

    // Update timestamp
    emqc_tm_update_timestamp(param);

    // log_info("payload: %s", json_object_to_json_string(jso));
    return jso;
}

json_object *emqc_tm_add_event_param(json_object *jso, emqc_tm_object *obj, bool ts)
{
    assert(jso && obj);
    char *key = obj->key;

    json_object *params = NULL;
    json_object *param  = NULL;
    json_object *val    = NULL;
    // check if params is exist
    if (!json_object_object_get_ex(jso, "params", &params)) {
        params = json_object_new_object();
        json_object_object_add(jso, "params", params);
    }

    if (ts) {

        if (!json_object_object_get_ex(params, "value", &val)) {
            val = json_object_new_object();
            json_object_object_add(params, "value", val);
        }

        if (!json_object_object_get_ex(val, key, &param)) {
            emqc_tm_add_param(val, key, obj);
            // emqc_tm_add_param(val, key, value, etp);
        }

        // Update timestamp
        emqc_tm_update_timestamp(params);

    } else {

        if (!json_object_object_get_ex(params, key, &param)) {
            emqc_tm_add_param(params, key, obj);
        }
    }

    // log_info("payload: %s", json_object_to_json_string(jso));
    return jso;
}

static void emqc_tm_property_array_set(json_object *jso_val, emqc_tm_object property_array[], int i)
{
    array_list *al = json_object_get_array(jso_val);
    size_t len = array_list_length(al);
    switch (property_array[i].type) {
        case EMQC_TM_ARR_INT:;
            int **value_int_arr = (int **) property_array[i].val_o;
            for (int j = 0; j < len; j++) {
                json_object *jso_i = (json_object *) array_list_get_idx(al, j);
                if (json_type_int == json_object_get_type(jso_i)) {
                    int val_int = json_object_get_int(jso_i);
                    *value_int_arr[j] = val_int;
                }
            }
            break;
        case EMQC_TM_ARR_STR:;
            char **value_str_arr = (char **) property_array[i].val_o;
            for (int j = 0; j < len; j++) {
                json_object *jso_i = (json_object *) array_list_get_idx(al, j);
                if (json_type_string == json_object_get_type(jso_i)) {
                    const char *val_string = json_object_get_string(jso_i);
                    memset(value_str_arr[j], 0, strlen(value_str_arr[j]));
                    sprintf(value_str_arr[j], "%s", val_string);
                }
            }
            break;
        case EMQC_TM_ARR_DOUBLE:;
            double **value_dou_arr = (double **) property_array[i].val_o;
            for (int j = 0; j < len; j++) {
                json_object *jso_i = (json_object *) array_list_get_idx(al, j);
                if (json_type_double == json_object_get_type(jso_i)) {
                    double val_dou = json_object_get_int(jso_i);
                    *value_dou_arr[j] = val_dou;
                }
            }
            break;
        case EMQC_TM_ARR_STRUCT:;
            emqc_tm_object **value_struct_arr = (emqc_tm_object **) property_array[i].val_o;
            // log_info("======> key: %s", property_array[i].key);
            for (int j = 0; j < len; j++) {
                json_object *jso_i = (json_object *) array_list_get_idx(al, j);
                // log_info("======> jso: %s", json_object_to_json_string(jso_i));
                // log_info("======> key: %s", value_struct_arr[j]->key);
                emqc_tm_property_set(jso_i, value_struct_arr[j], NULL);
            }
            break;
        default:
            log_err("Not support now!");
            break;
    }
}


void emqc_tm_property_set(json_object *jso_tmp, emqc_tm_object property_array[], char *model_id) 
{
    int i = 0;
    do {
        if (property_array[i].type == EMQC_TM_TERMINATOR) {
            break;
        }

        json_object *jso_val = NULL;
        char key[32];

        if (model_id) {
            snprintf(key, sizeof(key), "%s:%s", model_id, property_array[i].key);
        } else {
            snprintf(key, sizeof(key), "%s", property_array[i].key);
        }

        if (json_object_object_get_ex(jso_tmp, key, &jso_val)) {
            switch (property_array[i].type) {
            case EMQC_TM_STR:;
                const char *val_str = json_object_get_string(jso_val);
                memset(property_array[i].val_s, 0, strlen(property_array[i].val_s));
                sprintf(property_array[i].val_s, "%s", val_str);
                break;
            case EMQC_TM_DOUBLE:;
                double val_dou = json_object_get_double(jso_val);
                property_array[i].val_d = val_dou;
                break;
            case EMQC_TM_INT:;
                int val = json_object_get_int(jso_val);
                property_array[i].val_i = val;
                break;
            case EMQC_TM_STRUCT:;
                emqc_tm_object *param = (emqc_tm_object*)property_array[i].val_o;
                emqc_tm_property_set(jso_val, param, NULL);
                break;
            case EMQC_TM_ARR_INT:
            case EMQC_TM_ARR_STR:
            case EMQC_TM_ARR_DOUBLE:
            case EMQC_TM_ARR_STRUCT:
                emqc_tm_property_array_set(jso_val, property_array, i);
                break;
            default:
                log_err("Not support now!");
                break;
            }
        }
    } while (++i);
}
